import json
import click
from .uploader import Uploader
from .app import App
from . import config, auth


@click.group()
def main():
    pass


@main.command()
@click.argument("path")
@click.argument("bucket_id", type=str)
@click.option("--service-account-json", type=click.Path(exists=False))
def upload(path, bucket_id, service_account_json):
    u = Uploader(
        bucket_id=bucket_id,
        service_account_json=service_account_json,
    )
    u.start(path)


@main.command()
@click.option("--client-secrets-json", type=click.Path(exists=False), default=None)
def authenticate(client_secrets_json):
    json = client_secrets_json if client_secrets_json else config.client_secrets_path
    credentials = auth.get_credentials(json)
    auth.save_credentials(credentials, config.user_credentials_path)


@main.command()
@click.argument("path")
def gen_key(path):
    key = auth.gen_key(path)
    with open(path, "wb") as f:
        f.write(key)


@main.command("pickle-service-account-file")
@click.argument("path-in")
@click.argument("key-file")
@click.option("--output-file", default=config.service_account_path)
@click.option("--bucket", default=None)
def pickle_service_account_file(path_in, key_file, output_file, bucket):
    with open(path_in, "r") as f:
        data = json.load(f)
    auth.pickle_service_account_credentials(data, output_file, key_file, bucket)


@main.command("pickle-client-secrets-file")
@click.argument("path-in")
@click.argument("key-file")
def pickle_client_secrets_file(path_in, key_file):
    with open(path_in, "r") as f:
        data = json.load(f)
    auth.pickle_client_secret_file(data, config.client_secrets_path, key_file)


@main.command()
@click.argument("key-file", type=click.Path(exists=False))
@click.option("--title", default="GCS Uploader")
@click.option("--with-authentication", is_flag=True)
@click.option("--with-service-account", is_flag=True)
@click.option("--with-bucket-id", is_flag=True)
@click.option("--dry-run", is_flag=True)
def gui(
    key_file, title, with_authentication, with_service_account, with_bucket_id, dry_run
):
    app = App(
        key=key_file,
        title=title,
        with_authentication=with_authentication,
        with_service_account=with_service_account,
        with_bucket_id=with_bucket_id,
        dry_run=dry_run,
        log_file_path=None,
    )
    app.run()



from dataclasses import dataclass, field
import os
import threading
import pathlib
import time
from google.cloud import storage
import concurrent.futures as cf
from loguru import logger
from . import auth
from . import exceptions
from . import utils


@dataclass
class UploadStatus:
    filepath: pathlib.Path
    time_started: int = field(default_factory=time.time)
    time_complete: int = 0
    size: int = 0
    bytes_so_far: int = 0
    total_bytes: int = 0
    num_uploaded: int = 0
    num_files: int = 0
    num_errors: int = 0
    num_skipped: int = 0
    complete: bool = False

    def done(self):
        self.time_complete = time.time()
        self.complete = True

    @property
    def elapsed_time(self):
        return time.time() - self.time_started


class Uploader:
    def __init__(
        self,
        bucket_id=None,
        service_account_file=None,
        user_credentials_path=None,
        key=None,
        client_secrets_path=None,
        dry_run=False,
        prefix=None,
    ):

        self._stop = False
        self._thread = None
        self.dry_run = dry_run
        self.futures = None
        self.lock = threading.Lock()

        if prefix:
            self.prefix = prefix.rstrip("/") + "/"
        else:
            self.prefix = ""

        if bucket_id is not None and len(bucket_id) < 3:
            raise ValueError("Must provide a valid bucket ID")
        self.bucket_id = bucket_id

        self.client = None

        service_account_file = (
            pathlib.Path(service_account_file)
            if service_account_file
            else pathlib.Path("")
        )

        credentials = None

        service_account_file = (
            pathlib.Path(service_account_file)
            if service_account_file is not None
            else None
        )
        user_credentials_path = (
            pathlib.Path(user_credentials_path)
            if user_credentials_path is not None
            else None
        )
        client_secrets_path = (
            pathlib.Path(client_secrets_path)
            if client_secrets_path is not None
            else None
        )
        try:
            if (
                isinstance(user_credentials_path, pathlib.Path)
                and user_credentials_path.is_file()
            ):
                logger.debug(f"Using {user_credentials_path}")
                credentials = auth.unpickle_user_credentials(user_credentials_path, key)
            elif (
                isinstance(service_account_file, pathlib.Path)
                and service_account_file.is_file()
            ):
                logger.debug(f"Using {service_account_file}")
                credentials, bucket = auth.unpickle_service_account_credentials(
                    service_account_file, key, True
                )
                self.bucket_id = bucket if self.bucket_id is None else self.bucket_id
            elif (
                isinstance(client_secrets_path, pathlib.Path)
                and client_secrets_path.is_file()
            ):
                logger.debug(f"Running auth flow with {client_secrets_path}")
                credentials = self.run_oauth_flow(
                    user_credentials_path, client_secrets_path, key
                )
            self.client = storage.Client(project=None, credentials=credentials)
        except Exception:
            logger.exception("Authentication error")
            raise ValueError("Authentication failed")

    @staticmethod
    def run_oauth_flow(user_credentials_path, client_secrets_path, key):
        credentials = auth.get_credentials_via_client_secrets_file(
            client_secrets_path, key
        )
        auth.pickle_user_credentials(credentials, user_credentials_path, key)
        return credentials

    def start(
        self,
        paths,
        callback=lambda x: None,
    ):

        if not (paths and all([pathlib.Path(_).exists() for _ in paths])):
            raise ValueError("Must provide valid paths to files or folders")

        self._stop = False

        try:
            bucket = self.client.bucket(self.bucket_id)
        except Exception:
            logger.exception("Failed to instantiate bucket")
            raise Exception("Failed to connect as invalid destination URI")

        self._thread = threading.Thread(
            target=self._future_handler,
            args=(bucket, paths, callback),
            daemon=True,
        )
        self._thread.start()

    def _future_handler(self, bucket, paths, callback):

        self.futures = {}

        job = self._mock_upload if self.dry_run else self._upload

        self._upload_status = UploadStatus(None)

        max_workers = max(os.cpu_count() - 1, 2)
        with cf.ThreadPoolExecutor(max_workers=max_workers) as executor:

            for path, dest in utils.parse_paths(paths):
                if self._stop:
                    break

                size = path.stat().st_size

                # Start accumulating sizes for running stats
                with self.lock:
                    self._upload_status.total_bytes += size
                    self._upload_status.num_files += 1

                self.futures[
                    executor.submit(
                        job,
                        path,
                        (pathlib.Path(self.prefix) / dest).as_posix(),
                        bucket,
                        callback,
                    )
                ] = path

        self._upload_status.done()
        callback(self._upload_status)

    @property
    def active(self):
        if self._thread is not None and self._thread.is_alive():
            return True
        return False

    def stop(self):
        self._stop = True
        if self.futures:
            [_.cancel() for _ in self.futures.keys() if not _.done()]

    def _upload(self, path, dest, bucket, cb):
        if self._stop:
            return
        skipped = num_errors = 0
        try:
            blob = bucket.blob(str(dest))
            blob.upload_from_filename(str(path), checksum="crc32c")
            logger.info(f"Uploaded {path} to {dest}")
        except Exception as e:
            if exceptions.is_delete_permission_error(e):
                logger.warning(f"{path} already exists on the remote")
                skipped = 1
            else:
                logger.exception(f"Error attempting uploading {path}")
                num_errors = 1

        with self.lock:
            if skipped or num_errors:
                self._upload_status.num_skipped += skipped
                self._upload_status.num_errors += num_errors
            else:
                self._upload_status.num_uploaded += 1
                self._upload_status.bytes_so_far += path.stat().st_size
            self._upload_status.filepath = path
            cb(self._upload_status)

    def _mock_upload(self, path, dest, bucket, cb):
        if self._stop:
            return
        time.sleep(1)
        with self.lock:
            self._upload_status.num_uploaded += 1
            self._upload_status.bytes_so_far += path.stat().st_size
            cb(self._upload_status)



from cryptography.fernet import Fernet
import pickle
import os
import pathlib
from google_auth_oauthlib import flow
from google.oauth2 import service_account
from google.oauth2.credentials import Credentials


def get_credentials(client_id, client_secret):
    client_config = {
        "installed": {
            "client_id": client_id,
            "client_secret": client_secret,
            "redirect_uris": ["urn:ietf:wg:oauth:2.0:oob", "http://localhost"],
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
        }
    }
    scopes = ("https://www.googleapis.com/auth/devstorage.read_write",)

    app_flow = flow.InstalledAppFlow.from_client_config(client_config, scopes=scopes)
    app_flow.run_local_server()  # Should catch exception here?
    return app_flow.credentials


def credentials_to_dict(credentials):
    return {
        "token": credentials.token,
        "refresh_token": credentials.refresh_token,
        "token_uri": credentials.token_uri,
        "client_id": credentials.client_id,
        "client_secret": credentials.client_secret,
        "scopes": credentials.scopes,
    }


def pickle_user_credentials(credentials, path, key):
    path = pathlib.Path(path)
    if not path.parent.is_dir():
        path.parent.mkdir(parents=True, exist_ok=True)
    credentials.token = None
    data = credentials_to_dict(credentials)
    encrypt_fields(data, ("refresh_token", "client_id", "client_secret"), key)
    with open(path, "wb") as f:
        pickle.dump(data, f)


def unpickle_user_credentials(path, key):
    with open(path, "rb") as f:
        data = pickle.load(f)
    decrypt_fields(data, ("refresh_token", "client_id", "client_secret"), key)
    return Credentials(**data)


def encrypt_fields(data, keys, key_path):
    cipher = get_cipher(key_path)
    for key in keys:
        data[key] = cipher.encrypt(data[key].encode("utf-8"))


def decrypt_fields(data, keys, key_path):
    cipher = get_cipher(key_path)
    for key in keys:
        data[key] = cipher.decrypt(data[key]).decode("utf-8")


def unpickle_service_account_credentials(path, key, as_credentials=True):
    with open(path, "rb") as f:
        data = pickle.load(f)
    decrypt_fields(data, ("private_key_id", "private_key"), key)
    bucket = data.pop("bucket") if "bucket" in data else None
    if as_credentials:
        data = service_account.Credentials.from_service_account_info(data)
    return data, bucket


def pickle_service_account_credentials(data, path, key, bucket=None):
    data = {
        k: data[k]
        for k in [
            "private_key",
            "private_key_id",
            "token_uri",
            "client_email",
        ]
    }
    encrypt_fields(data, ("private_key_id", "private_key"), key)
    if bucket is not None:
        data["bucket"] = bucket
    with open(path, "wb") as f:
        pickle.dump(data, f)


def unpickle_client_secret_file(path, key):
    with open(path, "rb") as f:
        data = pickle.load(f)

    decrypt_fields(
        data["installed"],
        ("client_secret",),
        key,
    )
    return data


def pickle_client_secret_file(data, path, key):
    encrypt_fields(data["installed"], ("client_secret",), key)
    with open(path, "wb") as f:
        pickle.dump(data, f)


def get_credentials_via_client_secrets_file(pickled_client_secret_file, key):
    data = unpickle_client_secret_file(pickled_client_secret_file, key)
    appflow = flow.InstalledAppFlow.from_client_config(
        data,
        scopes=["https://www.googleapis.com/auth/devstorage.read_write"],
    )
    appflow.run_local_server()
    return appflow.credentials


def gen_key(path):
    return Fernet.generate_key()


def get_cipher(key):
    if os.path.isfile(key):
        with open(key, "rb") as f:
            return Fernet(f.read())
    return Fernet(key)


def default_bucket_from_service_account_file(service_account_path):
    if not os.path.isfile(service_account_path):
        return None
    with open(service_account_path, "rb") as f:
        data = pickle.load(f)
    return data.pop("bucket") if "bucket" in data else None
